; i8080 emulator
;   based on avrcpm project

.define ENABLE_IO

;.define BDOS
;.define TRACE
;.define TRACE_LINE
;.define DELAY

; i8080 cpu registers
.def rl   = r10
.def rh   = r11
.def re   = r12
.def rd   = r13
.def rc   = r14
.def rb   = r15
.def rf   = r18
.def ra   = r19
.def rspl = r20
.def rsph = r21
; X contains PC + offset
.def opl  = r24	; operand register low
.def oph  = r25	; operand register high

; emulator registers
.def reg00 = r4
.def reg04 = r5
.def reg10 = r6
.def exec  = r8
.def store = r9
.def state = r22

; bit numbers and masks
.equ flag_s = 7
.equ flag_z = 6
.equ flag_a = 4
.equ flag_p = 2
.equ flag_c = 0

.equ state_hlt = 0
.equ state_ie  = 1

.cseg

.ifdef TRACE
	.ifdef TRACE_LINE
	msg_trace_start:	.db 0, 0
	msg_trace_stop:		.db 0, 0
	msg_trace_line:
		.db 10, ">> [PC %X] [SP %X] [AF %X] "
		.db "[BC %X] [DE %X] [HL %X] [MEM %x %x]", 0

	.else
	msg_trace_start:	.db "[s[25;1H[?25l[7m", 0, 0
	msg_trace_stop:		.db "[K[0m[u[?25h", 0, 0
	msg_trace_line:
		.db 13, ">> [PC %X] [SP %X] [AF %X] "
		.db "[BC %X] [DE %X] [HL %X] [MEM %x %x]", 0
	.endif
.endif

.macro bmov	; to_reg, to_bit, from_reg, from_bit
	bst	@2, @3
	bld	@0, @1
.endmacro

; === INITIALIZE CPU AND START EXECUTION ==================================
;   start execution with r16:r17 offset
cpu_init:
	clr	reg00		; reg00 = 0x00
	ldi	opl, 0x04
	mov	reg04, opl	; reg04 = 0x04
	ldi	opl, high(MEM_BEGIN)
	mov	reg10, opl	; reg10 = 0x10
	ldi	opl, 0xFF
	ldi	oph, 0xFF
	movw	rf, opl		; AF = 0xFFFF
	movw	rc, opl		; BC = 0xFFFF
	movw	re, opl		; DE = 0xFFFF
	movw	rl, opl		; HL = 0xFFFF
	adiw	opl, 1
	movw	rspl, opl	; SP = 0x0000
	movw	xl, opl
	add	xh, reg10	; PC = 0x0000 + offset
	add	xl, r16
	adc	xh, r17		; PC = 0x0000 + r16:r17 + offset

	clr	state

; === EMULATE NEXT INSTRUCTION ============================================
cpu_instr:
	.ifdef TRACE
		push	zl
		push	zh

		; start trace line
		ldi	zl, low(msg_trace_start * 2)
		ldi	zh, high(msg_trace_start * 2)
		rcall	uart_printf

		; trace execution
		ld	zl, X+
		ld	zh, X+
		sbiw	xl, 2
		push	zh		; (PC + 1)
		push	zl		; (PC)

		push	rl
		push	rh		; HL
		push	re
		push	rd		; DE
		push	rc
		push	rb		; BC
		push	rf
		push	ra		; AF
		push	rspl
		push	rsph		; SP
		sub	xh, reg10
		push	xl
		push	xh		; PC (offset-adjusted)
		add	xh, reg10

		ldi	zl, low(msg_trace_line * 2)
		ldi	zh, high(msg_trace_line * 2)
		rcall	uart_printf

		; end trace line
		ldi	zl, low(msg_trace_stop * 2)
		ldi	zh, high(msg_trace_stop * 2)
		rcall	uart_printf

		.ifdef DELAY
			; delay
			clr	r0
			trace_delay:
			ldi	zl, 0xFF
			ldi	zh, 0x1F
			trace_delay_inner:
				sbiw	zl, 1
				brne	trace_delay_inner
			dec	r0
			brne	trace_delay
		.endif

		pop	zh
		pop	zl
	.endif

	; TODO: interrupt handling
	sbrc	state, state_hlt	;1
	ret				;4

	.ifdef BDOS
		push	r16
		cp	xh, reg10
		brne	no_bdos
		cpi	xl, 5
		brne	no_bdos		; check if PC = 0x0005
		mov	r16, rc
		cpi	r16, 2
		breq	bdos_conout	; func 02 = CONOUT
		cpi	r16, 9
		breq	bdos_printstr	; func 09 = PRINTSTR
		rjmp	no_bdos

		; print character
		bdos_conout:
			mov	r16, re
			rcall	uart_raw_putc
			rjmp	no_bdos

		; print string terminated by '$'
		bdos_printstr:
			mov	yl, re
			mov	yh, rd
			add	yh, reg10
		bdos_printstr_loop:
			ld	r16, Y+
			cpi	r16, '$'
			breq	bdos_printstr_done
			rcall	uart_raw_putc
			rjmp	bdos_printstr_loop

		bdos_printstr_done:

		no_bdos:
		pop	r16
	.endif

	; fetch opcode
	ld	opl, X+			;2

	; get instr table offset
	ldi	zl, low(instr_table * 2)	;1
	ldi	zh, high(instr_table * 2)	;1
	mul	opl, reg04			;2
	add	zl, r0				;1
	adc	zh, r1				;1

	; fetch values
	lpm	store, Z+		;3 (store)
	lpm	exec, Z+		;3 (exec)
	lpm	zl, Z+			;3 (fetch)
	ldi	zh, high(fetch_begin)	;1
	ijmp				;2 (fire!)

; === PHASE END MACROS ====================================================
.macro fetch_done
	ldi	zh, high(exec_begin)	;1
	mov	zl, exec		;1
	ijmp				;2
.endmacro

.macro exec_done
	ldi	zh, high(store_begin)	;1
	mov	zl, store		;1
	ijmp				;2
.endmacro

.macro instr_done
	rjmp	cpu_instr		;2
.endmacro

; === FETCH OPERAND =======================================================
.org (PC + 255) & 0xFF00
fetch_begin:
fetch_nop:
	fetch_done

fetch_a:
	mov opl, ra
	fetch_done

fetch_b:
	mov opl, rb
	fetch_done

fetch_c:
	mov opl, rc
	fetch_done

fetch_d:
	mov opl, rd
	fetch_done

fetch_e:
	mov opl, re
	fetch_done

fetch_h:
	mov opl, rh
	fetch_done

fetch_l:
	mov opl, rl
	fetch_done

fetch_af:
	movw opl, rf
	fetch_done

fetch_bc:
	movw opl, rc
	fetch_done

fetch_de:
	movw opl, re
	fetch_done

fetch_hl:
	movw opl, rl
	fetch_done

fetch_sp:
	movw opl, rspl
	fetch_done

fetch_mbc:
	movw yl, rc	;1
	add  yh, reg10	;1
	ld   opl, Y	;2
	fetch_done

fetch_mde:
	movw yl, re	;1
	add  yh, reg10	;1
	ld   opl, Y	;2
	fetch_done

fetch_mhl:
	movw yl, rl	;1
	add  yh, reg10	;1
	ld   opl, Y	;2
	fetch_done

fetch_msp:
	movw yl, rspl	;1
	add  yh, reg10	;1
	ld   opl, Y+	;2
	ld   oph, Y	;2
	fetch_done

fetch_dir8:
	ld   opl, X+	;2
	fetch_done

fetch_dir16:
	ld   opl, X+	;2
	ld   oph, X+	;2
	fetch_done

fetch_rst:
	sbiw xl, 1	;2
	ld   opl, X+	;2	re-load opcode
	andi opl, 0x38	;1
	clr  oph	;1
	fetch_done

fetch_end:
.if (high(fetch_end) != high(fetch_begin))
	.error "PROBLEM: fetch phase spans boundary, won't work"
.endif

; === STORE OPERAND =======================================================
;.org (PC + 255) & 0xFF00
store_begin:
store_nop:
	instr_done

store_a:
	mov ra, opl
	instr_done

store_b:
	mov rb, opl
	instr_done

store_c:
	mov rc, opl
	instr_done

store_d:
	mov rd, opl
	instr_done

store_e:
	mov re, opl
	instr_done

store_h:
	mov rh, opl
	instr_done

store_l:
	mov rl, opl
	instr_done

store_af:
	movw rf, opl	;1	load AF register
	andi rf, 0xD7	;1	clear bit5 and bit3
	ori  rf, 0x02	;1	set bit1
	instr_done

store_bc:
	movw rc, opl
	instr_done

store_de:
	movw re, opl
	instr_done

store_hl:
	movw rl, opl
	instr_done

store_sp:
	movw rspl, opl
	instr_done

store_pc:
	movw xl, opl
	add  xh, reg10
	instr_done

store_ma:
	movw yl, opl	;1
	add  yh, reg10	;1
	brcs store_nop	;1
	st   Y,  ra	;2
	instr_done

store_mbc:
	movw yl, rc
	add  yh, reg10
	brcs store_nop
	st   Y,  opl
	instr_done

store_mde:
	movw yl, re
	add  yh, reg10
	brcs store_nop
	st   Y,  opl
	instr_done

store_mhl:
	movw yl, rl
	add  yh, reg10
	brcs store_nop
	st   Y, opl
	instr_done

store_msp:
	movw yl, rspl	;1
	add  yh, reg10	;1
	brcs store_nop	;1
	st   Y+, opl	;2
	st   Y, oph	;2
	instr_done

store_ret:
	movw yl, rspl	;1
	add  yh, reg10	;1	Y = SP + offset
	ld   xl, Y+	;2
	ld   xh, Y+	;2	X = new PC
	add  xh, reg10	;1	X = new PC + offset
	sub  yh, reg10	;1
	movw rspl, yl	;1	update SP
	instr_done

store_call:
	movw yl, rspl	;1
	add  yh, reg10	;1	Y = SP + offset
	sub  xh, reg10	;1	X = PC
	st   -Y, xh	;2
	st   -Y, xl	;2
	sub  yh, reg10	;1
	movw rspl, yl	;1	update SP
	movw xl, opl	;1	X = new PC
	add  xh, reg10	;1	X = new PC + offset
	instr_done

store_end:
.if (high(store_end) != high(store_begin))
	.error "PROBLEM: store phase spans boundary, won't work"
.endif

; === MOVED EXECUTE OPERATIONS ============================================
.macro update_f
	ldi	zh, high(flag_table*2)	;1
	mov	zl, @0			;1
	lpm	rf, Z			;3
.endmacro

do_exec_inc:	; inc byte, affects S, Z, P, A
	mov  oph, rf			; save F
	inc  opl			;
	update_f opl			; set S, Z, P flags
	bmov rf, flag_c, oph, flag_c	; restore C flag

	mov  oph, opl			; save result
	andi oph, 0x0F			; mask low nibble
	brne do_exec_inc_done		; if non-zero ...
	ori  rf,  0x10			; ... set A flag
do_exec_inc_done:
	exec_done

do_exec_dec:	; dec byte, affects S, Z, P, A
	mov  oph, rf			; save F
	dec  opl			;
	update_f opl			; set S, Z, P flags
	bmov rf, flag_c, oph, flag_c	; restore C flag

	mov  oph, opl			; save result
	inc  oph			; allow zero-test
	andi oph, 0x0F			; mask low nibble
	breq do_exec_dec_done		; if zero ...
	ori  rf, 0x10			; ... set A flag
do_exec_dec_done:
	exec_done

do_exec_rla:	; rotate A left through carry, affects C
	clc				;1
	sbrc rf, flag_c			;1	CF -> C
	sec				;1
	bmov rf, flag_c, ra, 7		;2	bit 7 -> CF
	rol  ra				;1
	exec_done

do_exec_rra:	; rotate A right through carry, affects C
	clc				;1
	sbrc rf, flag_c			;1	CF -> C
	sec				;1
	bmov rf, flag_c, ra, 0		;2	bit 0 -> CF
	ror  ra				;1
	exec_done

do_exec_in:	; run IN callback
.ifdef ENABLE_IO
	mov   r17, opl			; r17 <- port
	rcall cpu_callback_in		; callback
	mov   ra, r16			; data -> r16
.else
	ldi   ra, 0xFF
.endif
	exec_done

do_exec_out:	; run OUT callback
.ifdef ENABLE_IO
	mov   r17, opl			; r17 <- port
	mov   r16, ra			; r16 <- data
	rcall cpu_callback_out		; callback
.endif
	exec_done

do_exec_daa:	; decimal adjust A, affects S, Z, P, A, C
	; temporaries
	ldi   opl, 0
	ldi   oph, 0

	; if (aux set) or (a & 0x0F > 9)
	;   ... add 0x06
	sbrc  rf,  flag_a
	ori   opl, 0x06
	mov   r17, ra
	andi  r17, 0x0F
	cpi   r17, 0x0A
	brlo  daa1
	ori   opl, 0x06

daa1:	; if (carry set) or (a >= 0x9A)
	sbrc  rf, flag_c
	rjmp  daa2a
	cpi   ra, 0x9A
	brlo  daa2b

daa2a:	; ... add 0x60 and set carry
	ori   opl, 0x60
	ori   oph, (1 << flag_c)

daa2b:	; do addition and update flags
	add   ra,  opl
	in    r17, sreg
	update_f ra
	bmov  rf, flag_a, r17, 5
	bmov  rf, flag_c, oph, flag_c

	exec_done

; === EXECUTE OPERATION ===================================================
.org (PC + 255) & 0xFF00
exec_begin:
exec_nop:	; no operation
	exec_done

exec_hlt:	; set HLT flag
	ori state, (1 << state_hlt)
	exec_done

exec_ei:	; set EI flag
	ori state, (1 << state_ie)
	exec_done

exec_di:	; clear EI flag
	andi state, ~(1 << state_ie)
	exec_done

exec_ina:	; moved
	rjmp do_exec_in

exec_outa:	; moved
	rjmp do_exec_out

exec_rmem8:	; read byte
	movw yl, opl	;1
	add  yh, reg10	;1
	ld   opl, Y	;2
	exec_done

exec_rmem16:	; read word
	movw yl, opl	;1
	add  yh, reg10	;1
	ld   opl, Y+	;2
	ld   oph, Y	;2
	exec_done

exec_push:	; write stack, update SP
	movw yl, rspl	;1
	add  yh, reg10	;1	Y = SP + offset
	st   -Y, oph	;2
	st   -Y, opl	;2
	sub  yh, reg10	;1	Y = SP - 2
	movw rspl, yl	;1	update SP
	exec_done

exec_pop:	; read stack, update SP
	movw yl, rspl	;1
	add  yh, reg10	;1	Y = SP + offset
	ld   opl, Y+	;2
	ld   oph, Y+	;2
	sub  yh, reg10	;1	Y = SP + 2
	movw rspl, yl	;1	update SP
	exec_done

exec_exhl:	; swap OP and HL
	movw yl, opl	;1
	movw opl, rl	;1
	movw rl, yl	;1
	exec_done

exec_sthl:	; store HL to memory
	movw yl, opl	;1
	add  yh, reg10	;1	Y = OP + offset
	brcs exec_nop	;1
	st   Y+, rl	;2
	st   Y,  rh	;2
	exec_done

exec_scf:	; set carry
	ori  rf, (1 << flag_c)
	exec_done

exec_ccf:	; complement carry
	ldi  yl, (1 << flag_c)
	eor  rf, yl
	exec_done

exec_ifz:	; cancel instr if NZ
	sbrs rf, flag_z
	instr_done
	exec_done

exec_ifnz:	; cancel instr if Z
	sbrc rf, flag_z
	instr_done
	exec_done

exec_ifc:	; cancel instr if NC
	sbrs rf, flag_c
	instr_done
	exec_done

exec_ifnc:	; cancel instr if C
	sbrc rf, flag_c
	instr_done
	exec_done

exec_ifm:	; cancel instr if NS
	sbrs rf, flag_s
	instr_done
	exec_done

exec_ifp:	; cancel instr if S
	sbrc rf, flag_s
	instr_done
	exec_done

exec_ifpe:	; cancel instr if NP
	sbrs rf, flag_p
	instr_done
	exec_done

exec_ifpo:	; cancel instr if P
	sbrc rf, flag_p
	instr_done
	exec_done

exec_inc16:	; inc word
	adiw opl, 1	;2
	exec_done

exec_dec16:	; dec word
	sbiw opl, 1	;2
	exec_done

exec_inc:	; moved
	rjmp do_exec_inc

exec_dec:	; dec byte, affects S, Z, P, A
	rjmp do_exec_dec

exec_addhl:	; add OP to HL, affects C
	andi rf, ~(1 << flag_c)	;	clear CF
	add  rl, opl		;
	adc  rh, oph		;	HL = HL + OP
	brcc exec_addhl_done	;
	ori  rf, (1 << flag_c)	;	set CF if needed
exec_addhl_done:
	exec_done

exec_adda:	; add A, affects S, Z, P, A, C
	add  ra, opl		;
	in   yl, sreg		;
	update_f ra		;
	bmov rf, flag_c, yl, 0	;	C -> CF
	bmov rf, flag_a, yl, 5	;	H -> AF
	exec_done

exec_adca:	; adc A, affects S, Z, P, A, C
	clc			;
	sbrc rf, flag_c		;	CF -> C
	sec			;
	adc  ra, opl		;
	in   yl, sreg		;
	update_f ra		;
	bmov rf, flag_c, yl, 0	;	C -> CF
	bmov rf, flag_a, yl, 5	;	H -> AF
	exec_done

exec_suba:	; sub A, affects S, Z, P, A, C
	sub  ra, opl		;
	in   yl, sreg		;
	update_f ra		;
	bmov rf, flag_c, yl, 0	;	C -> CF
	com  yl
	bmov rf, flag_a, yl, 5	;	H -> AF
	exec_done

exec_sbca:	; sbc A, affects S, Z, P, A, C
	clc			;
	sbrc rf, flag_c		;	put CF into SREG
	sec			;
	sbc  ra, opl		;
	in   yl, sreg		;
	update_f ra		;
	bmov rf, flag_c, yl, 0	;	C -> CF
	com  yl
	bmov rf, flag_a, yl, 5	;	H -> AF
	exec_done

exec_cpa:	; cmp A, affects S, Z, P, A, C
	mov  oph, ra		;
	sub  oph, opl		;
	in   yl, sreg		;
	update_f oph		;
	bmov rf, flag_c, yl, 0	;	C -> CF
	com  yl
	bmov rf, flag_a, yl, 5	;	H -> AF
	exec_done		;

exec_anda:	; and A, affects S, Z, P, A, C
	mov  oph, ra
	or   oph, opl
	andi oph, 0x08		;	8080 quirk: AF = (A | OP) & 0x08

	and ra, opl
	update_f ra

	bmov rf, flag_a, oph, 3	;	set AF if needed
	exec_done

exec_xora:	; xor A, affects S, Z, P, A, C
	eor ra, opl
	update_f ra
	exec_done

exec_ora:	; or A, affects S, Z, P, A, C
	or ra, opl
	update_f ra
	exec_done

exec_rlca:	; rotate A left, affects C
	bst  ra, 7		;1	bit 7 -> T
	lsl  ra			;1
	bld  ra, 0		;1	T -> bit 0
	bld  rf, flag_c		;1	T -> CF
	exec_done

exec_rrca:	; rotate A right, affects C
	bst  ra, 0		;1	bit 0 -> T
	lsr  ra			;1
	bld  ra, 7		;1	T -> bit 7
	bld  rf, flag_c		;1	T -> CF
	exec_done

exec_rla:	; moved
	rjmp do_exec_rla

exec_rra:	; moved
	rjmp do_exec_rra

exec_cpla:	; complement A
	com ra
	exec_done

exec_daa:	; moved
	rjmp do_exec_daa

exec_end:
.if (high(exec_end) != high(exec_begin))
	.error "PROBLEM: execution phase spans boundary, won't work"
.endif

.macro instr
	; table contains four-byte entries = 1024 bytes
	.db low(@2), low(@1), low(@0), 0
.endmacro

.org (PC+255) & 0xff00
instr_table:
	instr	fetch_nop,   exec_nop,   store_nop  ; 00       : NOP
	instr	fetch_dir16, exec_nop,   store_bc   ; 01 nn nn : LXI B, d16
	instr	fetch_a,     exec_nop,   store_mbc  ; 02       : STAX B
	instr	fetch_bc,    exec_inc16, store_bc   ; 03       : INX B
	instr	fetch_b,     exec_inc,   store_b    ; 04       : INR B
	instr	fetch_b,     exec_dec,   store_b    ; 05       : DCR B
	instr	fetch_dir8,  exec_nop,   store_b    ; 06 nn    : MVI B, d8
	instr	fetch_nop,   exec_rlca,  store_nop  ; 07       : RLC
	instr	fetch_nop,   exec_nop,   store_nop  ; 08       : *NOP
	instr	fetch_bc,    exec_addhl, store_nop  ; 09       : DAD B
	instr	fetch_mbc,   exec_nop,   store_a    ; 0a       : LDAX B
	instr	fetch_bc,    exec_dec16, store_bc   ; 0b       : DCX B
	instr	fetch_c,     exec_inc,   store_c    ; 0c       : INR C
	instr	fetch_c,     exec_dec,   store_c    ; 0d       : DCR C
	instr	fetch_dir8,  exec_nop,   store_c    ; 0e nn    : MVI C, d8
	instr	fetch_nop,   exec_rrca,  store_nop  ; 0f       : RRC

	instr	fetch_nop,   exec_nop,   store_nop  ; 10       : *NOP
	instr	fetch_dir16, exec_nop,   store_de   ; 11 nn nn : LXI D, d16
	instr	fetch_a,     exec_nop,   store_mde  ; 12       : STAX D
	instr	fetch_de,    exec_inc16, store_de   ; 13       : INX D
	instr	fetch_d,     exec_inc,   store_d    ; 14       : INR D
	instr	fetch_d,     exec_dec,   store_d    ; 15       : DCR D
	instr	fetch_dir8,  exec_nop,   store_d    ; 16 nn    : MVI D, d8
	instr	fetch_nop,   exec_rla,   store_nop  ; 17       : RAL
	instr	fetch_nop,   exec_nop,   store_nop  ; 18       : *NOP
	instr	fetch_de,    exec_addhl, store_nop  ; 19       : DAD D
	instr	fetch_mde,   exec_nop,   store_a    ; 1a       : LDAX D
	instr	fetch_de,    exec_dec16, store_de   ; 1b       : DCX D
	instr	fetch_e,     exec_inc,   store_e    ; 1c       : INR E
	instr	fetch_e,     exec_dec,   store_e    ; 1d       : DCR E
	instr	fetch_dir8,  exec_nop,   store_e    ; 1e nn    : MVI E, d8
	instr	fetch_nop,   exec_rra,   store_nop  ; 1f       : RAR

	instr	fetch_nop,   exec_nop,   store_nop  ; 20       : *NOP
	instr	fetch_dir16, exec_nop,   store_hl   ; 21 nn nn : LXI H, d16
	instr	fetch_dir16, exec_sthl,  store_nop  ; 22 nn nn : SHLD a16
	instr	fetch_hl,    exec_inc16, store_hl   ; 23       : INX H
	instr	fetch_h,     exec_inc,   store_h    ; 24       : INR H
	instr	fetch_h,     exec_dec,   store_h    ; 25       : DCR H
	instr	fetch_dir8,  exec_nop,   store_h    ; 26 nn    : MVI H, d8
	instr	fetch_nop,   exec_daa,   store_nop  ; 27       : DAA
	instr	fetch_nop,   exec_nop,   store_nop  ; 28       : *NOP
	instr	fetch_hl,    exec_addhl, store_nop  ; 29       : DAD H
	instr	fetch_dir16, exec_rmem16,store_hl   ; 2a nn nn : LHLD a16
	instr	fetch_hl,    exec_dec16, store_hl   ; 2b       : DCX H
	instr	fetch_l,     exec_inc,   store_l    ; 2c       : INR L
	instr	fetch_l,     exec_dec,   store_l    ; 2d       : DCR L
	instr	fetch_dir8,  exec_nop,   store_l    ; 2e nn    : MVI L, d8
	instr	fetch_nop,   exec_cpla,  store_nop  ; 2f       : CMA

	instr	fetch_nop,   exec_nop,   store_nop  ; 30       : *NOP
	instr	fetch_dir16, exec_nop,   store_sp   ; 31 nn nn : LXI SP, d16
	instr	fetch_dir16, exec_nop,   store_ma   ; 32 nn nn : STA a16
	instr	fetch_sp,    exec_inc16, store_sp   ; 33       : INX SP
	instr	fetch_mhl,   exec_inc,   store_mhl  ; 34       : INR M
	instr	fetch_mhl,   exec_dec,   store_mhl  ; 35       : DCR M
	instr	fetch_dir8,  exec_nop,   store_mhl  ; 36 nn    : MVI M, d8
	instr	fetch_nop,   exec_scf,   store_nop  ; 37       : STC
	instr	fetch_nop,   exec_nop,   store_nop  ; 38       : *NOP
	instr	fetch_sp,    exec_addhl, store_nop  ; 39       : DAD SP
	instr	fetch_dir16, exec_rmem8, store_a    ; 3a nn nn : LDA a16
	instr	fetch_sp,    exec_dec16, store_sp   ; 3b       : DCX SP
	instr	fetch_a,     exec_inc,   store_a    ; 3c       : INR A
	instr	fetch_a,     exec_dec,   store_a    ; 3d       : DCR A
	instr	fetch_dir8,  exec_nop,   store_a    ; 3e nn    : MVI A, d8
	instr	fetch_nop,   exec_ccf,   store_nop  ; 3f       : CMC

	instr	fetch_b,     exec_nop,   store_b    ; 40       : MOV B, B
	instr	fetch_c,     exec_nop,   store_b    ; 41       : MOV B, C
	instr	fetch_d,     exec_nop,   store_b    ; 42       : MOV B, D
	instr	fetch_e,     exec_nop,   store_b    ; 43       : MOV B, E
	instr	fetch_h,     exec_nop,   store_b    ; 44       : MOV B, H
	instr	fetch_l,     exec_nop,   store_b    ; 45       : MOV B, L
	instr	fetch_mhl,   exec_nop,   store_b    ; 46       : MOV B, M
	instr	fetch_a,     exec_nop,   store_b    ; 47       : MOV B, A
	instr	fetch_b,     exec_nop,   store_c    ; 48       : MOV C, B
	instr	fetch_c,     exec_nop,   store_c    ; 49       : MOV C, C
	instr	fetch_d,     exec_nop,   store_c    ; 4a       : MOV C, D
	instr	fetch_e,     exec_nop,   store_c    ; 4b       : MOV C, E
	instr	fetch_h,     exec_nop,   store_c    ; 4c       : MOV C, H
	instr	fetch_l,     exec_nop,   store_c    ; 4d       : MOV C, L
	instr	fetch_mhl,   exec_nop,   store_c    ; 4e       : MOV C, M
	instr	fetch_a,     exec_nop,   store_c    ; 4f       : MOV C, A

	instr	fetch_b,     exec_nop,   store_d    ; 50       : MOV D, B
	instr	fetch_c,     exec_nop,   store_d    ; 51       : MOV D, C
	instr	fetch_d,     exec_nop,   store_d    ; 52       : MOV D, D
	instr	fetch_e,     exec_nop,   store_d    ; 53       : MOV D, E
	instr	fetch_h,     exec_nop,   store_d    ; 54       : MOV D, H
	instr	fetch_l,     exec_nop,   store_d    ; 55       : MOV D, L
	instr	fetch_mhl,   exec_nop,   store_d    ; 56       : MOV D, M
	instr	fetch_a,     exec_nop,   store_d    ; 57       : MOV D, A
	instr	fetch_b,     exec_nop,   store_e    ; 58       : MOV E, B
	instr	fetch_c,     exec_nop,   store_e    ; 59       : MOV E, C
	instr	fetch_d,     exec_nop,   store_e    ; 5a       : MOV E, D
	instr	fetch_e,     exec_nop,   store_e    ; 5b       : MOV E, E
	instr	fetch_h,     exec_nop,   store_e    ; 5c       : MOV E, H
	instr	fetch_l,     exec_nop,   store_e    ; 5d       : MOV E, L
	instr	fetch_mhl,   exec_nop,   store_e    ; 5e       : MOV E, M
	instr	fetch_a,     exec_nop,   store_e    ; 5f       : MOV E, A

	instr	fetch_b,     exec_nop,   store_h    ; 60       : MOV H, B
	instr	fetch_c,     exec_nop,   store_h    ; 61       : MOV H, C
	instr	fetch_d,     exec_nop,   store_h    ; 62       : MOV H, D
	instr	fetch_e,     exec_nop,   store_h    ; 63       : MOV H, E
	instr	fetch_h,     exec_nop,   store_h    ; 64       : MOV H, H
	instr	fetch_l,     exec_nop,   store_h    ; 65       : MOV H, L
	instr	fetch_mhl,   exec_nop,   store_h    ; 66       : MOV H, M
	instr	fetch_a,     exec_nop,   store_h    ; 67       : MOV H, A
	instr	fetch_b,     exec_nop,   store_l    ; 68       : MOV L, B
	instr	fetch_c,     exec_nop,   store_l    ; 69       : MOV L, C
	instr	fetch_d,     exec_nop,   store_l    ; 6a       : MOV L, D
	instr	fetch_e,     exec_nop,   store_l    ; 6b       : MOV L, E
	instr	fetch_h,     exec_nop,   store_l    ; 6c       : MOV L, H
	instr	fetch_l,     exec_nop,   store_l    ; 6d       : MOV L, L
	instr	fetch_mhl,   exec_nop,   store_l    ; 6e       : MOV L, M
	instr	fetch_a,     exec_nop,   store_l    ; 6f       : MOV L, A

	instr	fetch_b,     exec_nop,   store_mhl  ; 70       : MOV M, B
	instr	fetch_c,     exec_nop,   store_mhl  ; 71       : MOV M, C
	instr	fetch_d,     exec_nop,   store_mhl  ; 72       : MOV M, D
	instr	fetch_e,     exec_nop,   store_mhl  ; 73       : MOV M, E
	instr	fetch_h,     exec_nop,   store_mhl  ; 74       : MOV M, H
	instr	fetch_l,     exec_nop,   store_mhl  ; 75       : MOV M, L
	instr	fetch_nop,   exec_hlt,   store_nop  ; 76       : HLT
	instr	fetch_a,     exec_nop,   store_mhl  ; 77       : MOV M, A
	instr	fetch_b,     exec_nop,   store_a    ; 78       : MOV A, B
	instr	fetch_c,     exec_nop,   store_a    ; 79       : MOV A, C
	instr	fetch_d,     exec_nop,   store_a    ; 7a       : MOV A, D
	instr	fetch_e,     exec_nop,   store_a    ; 7b       : MOV A, E
	instr	fetch_h,     exec_nop,   store_a    ; 7c       : MOV A, H
	instr	fetch_l,     exec_nop,   store_a    ; 7d       : MOV A, L
	instr	fetch_mhl,   exec_nop,   store_a    ; 7e       : MOV A, M
	instr	fetch_a,     exec_nop,   store_a    ; 7f       : MOV A, A

	instr	fetch_b,     exec_adda,  store_nop  ; 80       : ADD B
	instr	fetch_c,     exec_adda,  store_nop  ; 81       : ADD C
	instr	fetch_d,     exec_adda,  store_nop  ; 82       : ADD D
	instr	fetch_e,     exec_adda,  store_nop  ; 83       : ADD E
	instr	fetch_h,     exec_adda,  store_nop  ; 84       : ADD H
	instr	fetch_l,     exec_adda,  store_nop  ; 85       : ADD L
	instr	fetch_mhl,   exec_adda,  store_nop  ; 86       : ADD M
	instr	fetch_a,     exec_adda,  store_nop  ; 87       : ADD A
	instr	fetch_b,     exec_adca,  store_nop  ; 88       : ADC B
	instr	fetch_c,     exec_adca,  store_nop  ; 89       : ADC C
	instr	fetch_d,     exec_adca,  store_nop  ; 8a       : ADC D
	instr	fetch_e,     exec_adca,  store_nop  ; 8b       : ADC E
	instr	fetch_h,     exec_adca,  store_nop  ; 8c       : ADC H
	instr	fetch_l,     exec_adca,  store_nop  ; 8d       : ADC L
	instr	fetch_mhl,   exec_adca,  store_nop  ; 8e       : ADC M
	instr	fetch_a,     exec_adca,  store_nop  ; 8f       : ADC A

	instr	fetch_b,     exec_suba,  store_nop  ; 90       : SUB B
	instr	fetch_c,     exec_suba,  store_nop  ; 91       : SUB C
	instr	fetch_d,     exec_suba,  store_nop  ; 92       : SUB D
	instr	fetch_e,     exec_suba,  store_nop  ; 93       : SUB E
	instr	fetch_h,     exec_suba,  store_nop  ; 94       : SUB H
	instr	fetch_l,     exec_suba,  store_nop  ; 95       : SUB L
	instr	fetch_mhl,   exec_suba,  store_nop  ; 96       : SUB M
	instr	fetch_a,     exec_suba,  store_nop  ; 97       : SUB A
	instr	fetch_b,     exec_sbca,  store_nop  ; 98       : SBB B
	instr	fetch_c,     exec_sbca,  store_nop  ; 99       : SBB C
	instr	fetch_d,     exec_sbca,  store_nop  ; 9a       : SBB D
	instr	fetch_e,     exec_sbca,  store_nop  ; 9b       : SBB E
	instr	fetch_h,     exec_sbca,  store_nop  ; 9c       : SBB H
	instr	fetch_l,     exec_sbca,  store_nop  ; 9d       : SBB L
	instr	fetch_mhl,   exec_sbca,  store_nop  ; 9e       : SBB M
	instr	fetch_a,     exec_sbca,  store_nop  ; 9f       : SBB A

	instr	fetch_b,     exec_anda,  store_nop  ; a0       : ANA B
	instr	fetch_c,     exec_anda,  store_nop  ; a1       : ANA C
	instr	fetch_d,     exec_anda,  store_nop  ; a2       : ANA D
	instr	fetch_e,     exec_anda,  store_nop  ; a3       : ANA E
	instr	fetch_h,     exec_anda,  store_nop  ; a4       : ANA H
	instr	fetch_l,     exec_anda,  store_nop  ; a5       : ANA L
	instr	fetch_mhl,   exec_anda,  store_nop  ; a6       : ANA M
	instr	fetch_a,     exec_anda,  store_nop  ; a7       : ANA A
	instr	fetch_b,     exec_xora,  store_nop  ; a8       : XRA B
	instr	fetch_c,     exec_xora,  store_nop  ; a9       : XRA C
	instr	fetch_d,     exec_xora,  store_nop  ; aa       : XRA D
	instr	fetch_e,     exec_xora,  store_nop  ; ab       : XRA E
	instr	fetch_h,     exec_xora,  store_nop  ; ac       : XRA H
	instr	fetch_l,     exec_xora,  store_nop  ; ad       : XRA L
	instr	fetch_mhl,   exec_xora,  store_nop  ; ae       : XRA M
	instr	fetch_a,     exec_xora,  store_nop  ; af       : XRA A

	instr	fetch_b,     exec_ora,   store_nop  ; b0       : ORA B
	instr	fetch_c,     exec_ora,   store_nop  ; b1       : ORA C
	instr	fetch_d,     exec_ora,   store_nop  ; b2       : ORA D
	instr	fetch_e,     exec_ora,   store_nop  ; b3       : ORA E
	instr	fetch_h,     exec_ora,   store_nop  ; b4       : ORA H
	instr	fetch_l,     exec_ora,   store_nop  ; b5       : ORA L
	instr	fetch_mhl,   exec_ora,   store_nop  ; b6       : ORA M
	instr	fetch_a,     exec_ora,   store_nop  ; b7       : ORA A
	instr	fetch_b,     exec_cpa,   store_nop  ; b8       : CMP B
	instr	fetch_c,     exec_cpa,   store_nop  ; b9       : CMP C
	instr	fetch_d,     exec_cpa,   store_nop  ; ba       : CMP D
	instr	fetch_e,     exec_cpa,   store_nop  ; bb       : CMP E
	instr	fetch_h,     exec_cpa,   store_nop  ; bc       : CMP H
	instr	fetch_l,     exec_cpa,   store_nop  ; bd       : CMP L
	instr	fetch_mhl,   exec_cpa,   store_nop  ; be       : CMP M
	instr	fetch_a,     exec_cpa,   store_nop  ; bf       : CMP A

	instr	fetch_nop,   exec_ifnz,  store_ret  ; c0       : RNZ
	instr	fetch_nop,   exec_pop,   store_bc   ; c1       : POP B
	instr	fetch_dir16, exec_ifnz,  store_pc   ; c2 nn nn : JNZ a16
	instr	fetch_dir16, exec_nop,   store_pc   ; c3 nn nn : JMP a16
	instr	fetch_dir16, exec_ifnz,  store_call ; c4 nn nn : CNZ a16
	instr	fetch_bc,    exec_push,  store_nop  ; c5       : PUSH B
	instr	fetch_dir8,  exec_adda,  store_nop  ; c6 nn    : ADI d8
	instr	fetch_rst,   exec_nop,   store_call ; c7       : RST 0
	instr	fetch_nop,   exec_ifz,   store_ret  ; c8       : RZ
	instr	fetch_nop,   exec_nop,   store_ret  ; c9       : RET
	instr	fetch_dir16, exec_ifz,   store_pc   ; ca nn nn : JZ a16
	instr	fetch_dir16, exec_nop,   store_pc   ; cb nn nn : *JMP a16
	instr	fetch_dir16, exec_ifz,   store_call ; cc nn nn : CZ a16
	instr	fetch_dir16, exec_nop,   store_call ; cd nn nn : CALL a16
	instr	fetch_dir8,  exec_adca,  store_nop  ; ce nn    : ACI d8
	instr	fetch_rst,   exec_nop,   store_call ; cf       : RST 1

	instr	fetch_nop,   exec_ifnc,  store_ret  ; d0       : RNC
	instr	fetch_nop,   exec_pop,   store_de   ; d1       : POP D
	instr	fetch_dir16, exec_ifnc,  store_pc   ; d2 nn nn : JNC a16
	instr	fetch_dir8,  exec_outa,  store_nop  ; d3 nn    : OUT d8
	instr	fetch_dir16, exec_ifnc,  store_call ; d4 nn nn : CNC a16
	instr	fetch_de,    exec_push,  store_nop  ; d5       : PUSH D
	instr	fetch_dir8,  exec_suba,  store_nop  ; d6 nn    : SUI d8
	instr	fetch_rst,   exec_nop,   store_call ; d7       : RST 2
	instr	fetch_nop,   exec_ifc,   store_ret  ; d8       : RC
	instr	fetch_nop,   exec_nop,   store_ret  ; d9       : *RET
	instr	fetch_dir16, exec_ifc,   store_pc   ; da nn nn : JC a16
	instr	fetch_dir8,  exec_ina,   store_nop  ; db nn    : IN d8
	instr	fetch_dir16, exec_ifc,   store_call ; dc nn nn : CC a16
	instr	fetch_dir16, exec_nop,   store_call ; dd nn nn : *CALL a16
	instr	fetch_dir8,  exec_sbca,  store_nop  ; de nn    : SBI d8
	instr	fetch_rst,   exec_nop,   store_call ; df       : RST 3

	instr	fetch_nop,   exec_ifpo,  store_ret  ; e0       : RPO
	instr	fetch_nop,   exec_pop,   store_hl   ; e1       : POP H
	instr	fetch_dir16, exec_ifpo,  store_pc   ; e2 nn nn : JPO a16
	instr	fetch_msp,   exec_exhl,  store_msp  ; e3       : XTHL
	instr	fetch_dir16, exec_ifpo,  store_call ; e4 nn nn : CPO a16
	instr	fetch_hl,    exec_push,  store_nop  ; e5       : PUSH H
	instr	fetch_dir8,  exec_anda,  store_nop  ; e6 nn    : ANI d8
	instr	fetch_rst,   exec_nop,   store_call ; e7       : RST 4
	instr	fetch_nop,   exec_ifpe,  store_ret  ; e8       : RPE
	instr	fetch_hl,    exec_nop,   store_pc   ; e9       : PCHL
	instr	fetch_dir16, exec_ifpe,  store_pc   ; ea nn nn : JPE a16
	instr	fetch_de,    exec_exhl,  store_de   ; eb       : XCHG
	instr	fetch_dir16, exec_ifpe,  store_call ; ec nn nn : CPE a16
	instr	fetch_dir16, exec_nop,   store_call ; ed nn nn : *CALL a16
	instr	fetch_dir8,  exec_xora,  store_nop  ; ee nn    : XRI d8
	instr	fetch_rst,   exec_nop,   store_call ; ef       : RST 5

	instr	fetch_nop,   exec_ifp,   store_ret  ; f0       : RP
	instr	fetch_nop,   exec_pop,   store_af   ; f1       : POP PSW
	instr	fetch_dir16, exec_ifp,   store_pc   ; f2 nn nn : JP a16
	instr	fetch_nop,   exec_di,    store_nop  ; f3       : DI
	instr	fetch_dir16, exec_ifp,   store_call ; f4 nn nn : CP a16
	instr	fetch_af,    exec_push,  store_nop  ; f5       : PUSH PSW
	instr	fetch_dir8,  exec_ora,   store_nop  ; f6 nn    : ORI d8
	instr	fetch_rst,   exec_nop,   store_call ; f7       : RST 6
	instr	fetch_nop,   exec_ifm,   store_ret  ; f8       : RM
	instr	fetch_hl,    exec_nop,   store_sp   ; f9       : SPHL
	instr	fetch_dir16, exec_ifm,   store_pc   ; fa nn nn : JM a16
	instr	fetch_nop,   exec_ei,    store_nop  ; fb       : EI
	instr	fetch_dir16, exec_ifm,   store_call ; fc nn nn : CM a16
	instr	fetch_dir16, exec_nop,   store_call ; fd nn nn : *CALL a16
	instr	fetch_dir8,  exec_cpa,   store_nop  ; fe nn    : CPI d8
	instr	fetch_rst,   exec_nop,   store_call ; ff       : RST 7

; === FLAG TABLE (i8080) ==================================================
;   contains flags for S, Z, 5, 3, P, 1
.org (PC + 255) & 0xFF00
flag_table:
	.db 0x46, 0x02, 0x02, 0x06, 0x02, 0x06, 0x06, 0x02
	.db 0x02, 0x06, 0x06, 0x02, 0x06, 0x02, 0x02, 0x06
	.db 0x02, 0x06, 0x06, 0x02, 0x06, 0x02, 0x02, 0x06
	.db 0x06, 0x02, 0x02, 0x06, 0x02, 0x06, 0x06, 0x02
	.db 0x02, 0x06, 0x06, 0x02, 0x06, 0x02, 0x02, 0x06
	.db 0x06, 0x02, 0x02, 0x06, 0x02, 0x06, 0x06, 0x02
	.db 0x06, 0x02, 0x02, 0x06, 0x02, 0x06, 0x06, 0x02
	.db 0x02, 0x06, 0x06, 0x02, 0x06, 0x02, 0x02, 0x06
	.db 0x02, 0x06, 0x06, 0x02, 0x06, 0x02, 0x02, 0x06
	.db 0x06, 0x02, 0x02, 0x06, 0x02, 0x06, 0x06, 0x02
	.db 0x06, 0x02, 0x02, 0x06, 0x02, 0x06, 0x06, 0x02
	.db 0x02, 0x06, 0x06, 0x02, 0x06, 0x02, 0x02, 0x06
	.db 0x06, 0x02, 0x02, 0x06, 0x02, 0x06, 0x06, 0x02
	.db 0x02, 0x06, 0x06, 0x02, 0x06, 0x02, 0x02, 0x06
	.db 0x02, 0x06, 0x06, 0x02, 0x06, 0x02, 0x02, 0x06
	.db 0x06, 0x02, 0x02, 0x06, 0x02, 0x06, 0x06, 0x02
	.db 0x82, 0x86, 0x86, 0x82, 0x86, 0x82, 0x82, 0x86
	.db 0x86, 0x82, 0x82, 0x86, 0x82, 0x86, 0x86, 0x82
	.db 0x86, 0x82, 0x82, 0x86, 0x82, 0x86, 0x86, 0x82
	.db 0x82, 0x86, 0x86, 0x82, 0x86, 0x82, 0x82, 0x86
	.db 0x86, 0x82, 0x82, 0x86, 0x82, 0x86, 0x86, 0x82
	.db 0x82, 0x86, 0x86, 0x82, 0x86, 0x82, 0x82, 0x86
	.db 0x82, 0x86, 0x86, 0x82, 0x86, 0x82, 0x82, 0x86
	.db 0x86, 0x82, 0x82, 0x86, 0x82, 0x86, 0x86, 0x82
	.db 0x86, 0x82, 0x82, 0x86, 0x82, 0x86, 0x86, 0x82
	.db 0x82, 0x86, 0x86, 0x82, 0x86, 0x82, 0x82, 0x86
	.db 0x82, 0x86, 0x86, 0x82, 0x86, 0x82, 0x82, 0x86
	.db 0x86, 0x82, 0x82, 0x86, 0x82, 0x86, 0x86, 0x82
	.db 0x82, 0x86, 0x86, 0x82, 0x86, 0x82, 0x82, 0x86
	.db 0x86, 0x82, 0x82, 0x86, 0x82, 0x86, 0x86, 0x82
	.db 0x86, 0x82, 0x82, 0x86, 0x82, 0x86, 0x86, 0x82
	.db 0x82, 0x86, 0x86, 0x82, 0x86, 0x82, 0x82, 0x86

; vim: set ft=asm:
