; I/O HANDLERS
; vim: set ft=asm:

; registers
;   0x10 - TTY DATA (R/W)
;   0x11 - TTY STATUS (R/O)
;     0x00 - no data available
;     0xFF - data available
;   0x18 - SPI DATA   (R/W)
;   0x19 - SPI SELECT (W/O)
;   0x1A - SPI SPEED  (W/O)

.cseg

; called from CPU core
;   r17 - port (from CPU)
;   r16 - data (to   CPU)
; preserve all except r0-r3, r16-r17
.macro Mcpuin
	cpi	r17, 0x10	; port 0x10
	breq	ttyrd		; -> console data
	cpi	r17, 0x11	; port 0x11
	breq	ttyst		; -> console status
	cpi	r17, 0x18	; port 0x18
	breq	spird		; -> spi data
noport:	ldi	r16, 0xFF	; no response
	ret

; console functions
ttyst:	rcall	uart_has_data
	ret
ttyrd:	rcall	uart_has_data	; data available?
	cpi	r16, 0		; if not, return 0xFF
	breq	noport
	rcall	uart_raw_getc	; else get data
	ret

; spi functions
spird:	rcall	spi_getdata
	ret
.endm
cpu_callback_in:
	Mcpuin

; called from CPU core
;   r17 - port (from CPU)
;   r16 - data (from CPU)
; preserve all except r0-r3, r16-r17
.macro Mcpuout
	cpi	r17, 0x10	; port 0x10
	breq	ttywr		; -> console data
	cpi	r17, 0x18	; port 0x18
	breq	spiwr		; -> spi data
	cpi	r17, 0x19	; port 0x19
	breq	spisel		; -> spi select
	cpi	r17, 0x1A	; port 0x1A
	breq	spispd		; -> spi select
	ret

; console functions
ttywr:	rcall	uart_raw_putc
	ret

; spi functions
spiwr:	rcall	spi_setdata
	ret
spisel:	rcall	spi_setsel
	ret
spispd:	rcall	spi_setspd
	ret
.endm
cpu_callback_out:
	Mcpuout

