; SPI DRIVER
; vim: set ft=asm:

.equ SPI_CS	= PINB4
.equ SPI_MOSI	= PINB5
.equ SPI_MISO	= PINB6
.equ SPI_SCK	= PINB7

.cseg

; initialize SPI in master mode
spi_init:
	; CS, MOSI and SCK as outputs, MISO as input
	in	r16, DDRB
	ori	r16,  (1 << SPI_CS) | (1 << SPI_MOSI) | (1 << SPI_SCK)
	andi	r16, ~(1 << SPI_MISO)
	out	DDRB, r16

	; set CS high and enable pullup on MISO
	in	r16, PORTB
	ori	r16, (1 << SPI_CS) | (1 << SPI_MISO)
	out	PORTB, r16

	; enable SPI as master, slow clock
	ldi	r16, (1 << SPE) | (1 << MSTR) | (1 << SPR1) | (1 << SPR0)
	out	SPCR, r16
	ldi	r16, 0x00
	out	SPSR, r16

	; clock a few bytes
	ldi	r17, 4
	ldi	r16, 0xFF
spi_init1:
	out	SPDR, r16
spi_init2:
	sbis	SPSR, SPIF
	rjmp	spi_init2
	dec	r17
	brne	spi_init1

	ldi	r16, 0x00
	out	SPDR, r16
	ret

; read data byte
spi_getdata:
	in	r16, SPDR
	ret

; write data byte
spi_setdata:
	out	SPDR, r16
spi_setdatal:
        sbis    SPSR, SPIF
        rjmp    spi_setdatal
	ret

; select device
spi_setsel:
	in	r17, PORTB
	cpi	r16, 0x00		; disable?
	breq	spi_disable		; ..yes
	andi	r17, ~(1 << SPI_CS)	; ..no, enable (low)
	out	PORTB, r17
	ret
spi_disable:				; ..yes, disable (high)
	ori	r17, (1 << SPI_CS)
	out	PORTB, r17
	ret

; device speed -> 0x00 is slow (fOsc/128), else is fast (fOsc/2)
spi_setspd:
	cpi	r16, 0x00
	breq	spi_slow
	ldi	r16, (1 << SPE) | (1 << MSTR)
	out	SPCR, r16
	ldi	r16, (1 << SPI2X)
	out	SPSR, r16
	ret
spi_slow:
	ldi	r16, (1 << SPE) | (1 << MSTR) | (1 << SPR1) | (1 << SPR0)
	out	SPCR, r16
	ldi	r16, 0x00
	out	SPSR, r16
	ret
