; XMEM driver
; vim: set ft=asm:

; xmem initialization
xmem_init:
	ldi	r16, (1 << SRE) | (1 << SRW10)
	out	MCUCR, r16
	ldi	r16, (1 << SRW11)
	out	EMCUCR, r16
	ret

; xmem memory test
; trashes r0, r1, r16, r17, X, Y and memory
; T is set on error, clear otherwise
.macro Mmemtest
	ldi	r16, 3
	clt			; clear T

	ldi	xh, high(MEM_BEGIN)
	ldi	xl, low(MEM_BEGIN)
	ldi	yh, high(MEM_SIZE)
	ldi	yl, low(MEM_SIZE)
write:	mul	xl, r16		; r0:r1 = xl * 3
	st	X+, r0		; write r0
	sbiw	yl, 1		; one less byte
	brne	write

	ldi	xh, high(MEM_BEGIN)
	ldi	xl, low(MEM_BEGIN)
	ldi	yh, high(MEM_SIZE)
	ldi	yl, low(MEM_SIZE)
read:	mul	xl, r16		; r0:r1 = xl * 3
	ld	r17, X+		; read to r17
	cpse	r17, r0		; if not equal
	set			; set T
	sbiw	yl, 1		; one less byte
	brne	read

	ret
.endm
xmem_test:
	Mmemtest

; load IHEX file
; trashes r16-r21 and Z
; T is set on error, clear otherwise
.macro Mmemload
	; r18 = len
	; r19 = type
	; r20 = data
	; r21 = sum
	; Z   = offset

	; wait for RECORD START
wait:	rcall	uart_raw_getc
	cpi	r16, ':'
	brne	wait

	; clear checksum
	clr	r21

	; read LENGTH (byte) -> r18
	rcall	uart_gethex8
	mov	r18, r16
	add	r21, r16

	; read OFFSET (high-low) -> Z
	rcall	uart_gethex8
	mov	zh, r16
	add	r21, r16
	rcall	uart_gethex8
	mov	zl, r16
	add	r21, r16

	; read TYPE (byte) -> r19
	rcall	uart_gethex8
	mov	r19, r16
	add	r21, r16

	; print address for type 0 (data)
	cpi	r19, 0x00
	brne	adjust
	ldi	r16, 13
	rcall	uart_raw_putc
	movw	r16, zl
	rcall	uart_puthex16

	; set memory offset
adjust:	subi	zh, -(high(MEM_BEGIN))

	; read DATA
loop:	cpi	r18, 0		; if zero-length
	breq	cksum		; -> done

	dec	r18
	rcall	uart_gethex8
	add	r21, r16
	cpi	r19, 0x00	; if type not DATA
	brne	loop		; -> skip writing
	st	Z+, r16
	rjmp	loop

	; read CKSUM
cksum:	rcall	uart_gethex8
	add	r21, r16	; if checksum not zero
	brne	fail		; -> fail

	; check if EOF
	cpi	r19, 0x01	; if type not EOF
	brne	wait		; -> repeat

	; remove final '\r\n'
	rcall	uart_raw_getc	; read byte
	ldi	r17, 13
	cpse	r16, r17	; if not '\n'
	rcall	uart_raw_getc	; read another byte

	clt			; clear T flag
	rjmp	done		; finished

fail:	set			; set T flag
done:	ret
.endm
xmem_load:
	Mmemload
