; DISK DRIVE ROUTINES - SD CARD
; vim: set ft=asm:

; messages
SD_MSG0:  db "  SD CARD ", 0
SD_MSG1:  db "NOT "
SD_MSG2:  db "FOUND", 13, 10, 0

SD_MSG3:  db "CMD0:", 0
SD_MSG4:  db "CMD8:", 0
SD_MSG5:  db "ACMD41 V1:", 0
SD_MSG6:  db "V2 CARD NOT SUPPORTED", 13, 10, 0
SD_MSG7:  db "CMD16:", 0
SD_MSG8:  db "SD CARD READ ERROR CMD17:", 0
SD_MSG9:  db "SD CARD WRITE ERROR CMD24:", 0

SD_TIME2: db "SD CARD "
SD_TIME:  db "TIMEOUT", 13, 10, 0
SD_CRLF:  db 13, 10, 0

; commands
SD_CMD0:      db 0x40, 0x00, 0x00, 0x00, 0x00, 0x95
SD_CMD8:      db 0x48, 0x00, 0x00, 0x01, 0xAA, 0x87
SD_CMD16:     db 0x50, 0x00, 0x00, 0x02, 0x00, 0x01
SD_CMD17:     db 0x51, 0x00, 0x00, 0x00, 0x00, 0x01
SD_CMD24:     db 0x58, 0x00, 0x00, 0x00, 0x00, 0x01
SD_ACMD41_V1: db 0x69, 0x40, 0x00, 0x00, 0x00, 0x01
SD_ACMD41_V2: db 0x69, 0x00, 0x00, 0x00, 0x00, 0x01
SD_CMD55:     db 0x77, 0x00, 0x00, 0x00, 0x00, 0x01

; sd card initialization
sdinit:   ld   hl, SD_MSG0   ; print message
          call outstr
          xor  a
          out  (SPISEL), a   ; deselect card
          out  (SPISPD), a   ; slow speed
          dec  a             ; send 0xFF
          ld   b, 0x10       ; 16 times
sdinit1:  out  (SPID), a
          dec  b
          jp   nz, sdinit1
          in   a, (SPID)     ; read input
          cp   0xFF          ; must be high
          jp   z, sdinit2    ; ...ok, continue

          ld   hl, SD_MSG1   ; not found
          call outstr
          call sdoff
          halt               ; halt system

sdinit2:  ; send CMD0 (go idle)
          call sdon          ; select card
          ld   hl, SD_CMD0
          ld   b,  6
          call sdcmd
          jp   c,  sdtimeout ; timeout
          cp   0x01          ; idle, no error?
          jp   z, sdinit3    ; ...yes, continue
          ld   hl, SD_MSG3   ; CMD0 error
          jp   sdfatal

sdinit3:  ; send CMD8 (set interface)
          call sdoff         ; deselect card
          call sdon          ; select card
          ld   hl, SD_CMD8
          ld   b,  6
          call sdcmd
          jp   c, sdtimeout  ; timeout
          and  0xFE          ; mask busy flag
          cp   0x04          ; illegal command?
          jp   sdinit4a      ; ...yes, ver.1 card
          and  0xFE          ; any other errors?
          jp   z,  sdinit5a  ; ...no, ver.2 card
          ld   hl, SD_MSG4   ; CMD8 error
          jp   sdfatal

          ; ver.1 card initialization
          ; cards are always byte-addressed
sdinit4a: call sdoff         ; deselect card
          call sdon          ; select card
sdinit4b: ld   hl, SD_CMD55  ; ACMD prefix
          ld   b,  6
          call sdcmd
          jp   c,  sdtimeout ; timeout
          and  0xFE          ; any errors?
          jp   z,  sdinit4c  ; ...no, continue
          jp   sdinit4d      ; command error
sdinit4c: ld   hl, SD_ACMD41_V1
          ld   b,  6
          call sdcmd
          jp   c,  sdtimeout ; timeout
          or   a             ; initialization finished?
          jp   z,  sdinit6   ; ...yes, continue
          and  0xFE          ; ...no, any errors?
          jp   z,  sdinit4b  ; ...no, resend command
sdinit4d: ld   hl, SD_MSG5   ; ACMD41 (v1) error
          jp   sdfatal

          ; ver.2 card initialization (TODO)
          ; cards can be byte- or block-addressed
sdinit5a: ld   hl, SD_MSG6
          call outstr
          call sdoff
          halt

          ; send CMD16 (set block length)
sdinit6:  call sdoff         ; deselect card
          call sdon          ; select card
          ld   hl, SD_CMD16
          ld   b,  6
          call sdcmd
          jp   c,  sdtimeout ; timeout
          and  0xFE          ; any errors?
          jp   z, sdinit7    ; ...no, continue
          ld   hl, SD_MSG7   ; CMD16 error
          jp   sdfatal

sdinit7:  call sdoff         ; deselect card
          ld   a, 0xFF
          out  (SPISPD), a   ; fast speed
          ld   hl, SD_MSG2   ; card initialized successfully
          call outstr
          ret

sdtimeout: ld  hl, SD_TIME   ; timeout message
          call outstr
          call sdoff         ; deselect card
          halt               ; halt system

sdfatal:  push af            ; save error code
          call outstr        ; print message
          pop  af            ; restore error code
          call outhex8       ; and print
          ld   hl, SD_CRLF   ; print newline
          call outstr
          call sdoff         ; deselect card
          halt               ; halt system

; send command
;    input: HL = cmd addr
;           B  = cmd len
;   output: CF = set if timeout
;           A  = response byte
sdcmd:    ld   a, (hl)       ; read byte
          out  (SPID), a     ; send byte
          inc  hl
          dec  b
          jp   nz, sdcmd     ; until done
          ld   b,  8         ; timeout value
sdcmd1:   ld   a,  0xFF
          out  (SPID), a     ; send 0xFF
          in   a,  (SPID)    ; read response
          ld   c,  a
          and  0x80          ; bit7 clear?
          jp   z,  sdcmd2    ; ...yes, valid
          dec  b             ; timeout?
          jp   nc, sdcmd1    ; ...no, try again
          ret                ; return timeout
sdcmd2:   ld   a,  c         ; response in A
          ret

; select card
sdon:     ld   a, 0xFF
          out  (SPISEL), a   ; select card
          out  (SPID), a     ; send 0xFF
          ret

; deselect card
sdoff:    ld   a, 0xFF
          out  (SPID), a     ; send 0xFF
          inc  a
          out  (SPISEL), a   ; deselect card
          dec  a
          out  (SPID), a     ; send 0xFF
          out  (SPID), a     ; send 0xFF
          ret

; convert host (dsk/trk/sec) to lba block address
; 0000 0000 0000 00DD DDTT TTTT TTSS SSSS + 0x2000
;   => 256 tracks, 64 sec/trk, offset 4 MB
;   input: HL       = dst addr
;          (HSTDSK) = disk
;          (HSTTRK) = track
;          (HSTSEC) = sector
;  output: nothing
sdlbablk: inc  hl            ; start from low byte
          inc  hl
          inc  hl

          ; 4th byte: TTSS SSSS
          ld   a, (HSTTRK)   ; trk 7654 3210
          rrca               ; trk 0765 4321
          rrca               ; trk 1076 5432
          ld   c, a
          and  0xC0          ; trk 10-- ----
          ld   b, a
          ld   a, (HSTSEC)
          or   b             ; add sectors
          ld   (hl), a       ; write result
          dec  hl

          ; 3rd byte: DDTT TTTT (+ offset)
          ld   a, c          ; trk 1076 5432
          and  0x3F          ; trk --76 5432
          ld   b, a
          ld   a, (HSTDSK)   ; dsk ---- 3210
          and  0x0F          ; mask disk
          rrca               ; dsk 0--- -321
          rrca               ; dsk 10-- --32
          ld   c, a
          and  0xC0          ; dsk 10-- ----
          or   b             ; add tracks
          add  0x20          ; add offset (sets CF)
          ld   (hl), a       ; write result
          dec  hl            ; no flag changes

          ; 2nd byte: 0000 00DD (+ offset)
          ld   a, 0x00       ; no flag changes
          adc  a, c          ; dsk 10-- --32 (+CF)
          and  0x03          ; dsk ---- --32
          ld   (hl), a
          dec  hl

          ; 1st byte: 0000 0000
          xor  a
          ld   (hl), a
          ret

; convert host (dsk/trk/sec) to lba byte address
; see sdlbablk for details
; TODO: skip conversion for block-addressed cards
sdlba:    push hl
          call sdlbablk      ; set block address
          pop  hl
          inc  hl            ; start at low byte
          inc  hl
          inc  hl
          ld   a, (hl)       ; load 4th byte
          add  a, a          ; double
          ld   b, a          ; B = new 3rd byte
          dec  hl
          ld   a, (hl)       ; load 3rd byte
          adc  a, a          ; double + CF
          ld   c, a          ; C = new 2nd byte
          dec  hl
          ld   a, (hl)       ; load 2nd byte
          adc  a, a          ; double + CF
          dec  hl
          ld   (hl), a       ; write 1st byte
          inc  hl
          ld   (hl), c       ; write 2nd byte
          inc  hl
          ld   (hl), b       ; write 3rd byte
          inc  hl
          xor  a
          ld   (hl), a       ; write 4th byte zero
          ret

; disk initialization
;   input: nothing
;  output: (BOOTDISK) = boot drive
dskinit:  xor  a             ; always boot from A:
          ld   (BOOTDISK), a
          call sdinit        ; initialize card
          ret

; disk select
;   input: C         = drive
;  output: HL        = DPH or 0x0000 on error
;          (BOOTSPT) = sec/trk in system area
;          (SEKDSK)  = drive
;          (SEKDPH)  = DPH
dsksel:   ld   hl, 0x0000
          ld   a,  c
          cp   4             ; valid drive?
          ret  nc            ; no, return

          ; get drive DPH
          ld   (SEKDSK), a   ; save drive
          add  a             ; drive x 2
          add  a             ; drive x 4
          add  a             ; drive x 8
          add  a             ; drive x 16
          ld   b,  0
          ld   c,  a
          ld   hl, DPHTAB
          add  hl, bc        ; HL = drive DPH
          ld   (SEKDPH), hl  ; save dph

          ld   bc, 10
          add  hl, bc        ; HL = DPB pointer
          ld   a,  (hl)      ; dereference
          inc  hl
          ld   h,  (hl)
          ld   l,  a
          ld   a,  (hl)      ; A  = low(SPT)
          ld   (BOOTSPT), a
          ld   hl, (SEKDPH)
          ret

; disk status
;   input: A  = drive
;  output: ZF = set if drive ok
dskst:    cp   4             ; valid drive?
          jp   nc, dskerr2   ; ...no, fail
          or   a             ; ...yes, success
          ret

; disk read
;   input: nothing
;  output: A = 0x00 on success
;          data in DSKBUF
dskread:  ld   hl, SD_CMD17+1
          call sdlba         ; set LBA address

          ; read command
          call sdon          ; select card
          ld   hl, SD_CMD17  ; read block
          ld   b,  6
          call sdcmd
          jp   c,  dskerr2   ; timeout, error
          and  0xFE          ; any error?
          jp   z,  dskrd1    ; ...no, continue

          ld   hl, SD_MSG8   ; error
          jp   dskerr

          ; wait for card ready
dskrd1:   ld   a, 0xFF
          out  (SPID), a     ; send 0xFF
          in   a, (SPID)     ; read token
          cp   0xFE          ; data ready?
          jp   nz, dskrd1    ; ...no, wait more

          ; read sector
          ld   hl, DSKBUF
          ld   bc, 0x0200
dskrd2:   ld   a, 0xFF
          out  (SPID), a
          in   a, (SPID)
          ld   (hl), a
          inc  hl
          dec  bc
          ld   a, b
          or   c
          jp   nz, dskrd2

          ld   a, 0xFF
          out  (SPID), a     ; drop crc byte 1
          out  (SPID), a     ; drop crc byte 2

          call sdoff         ; deselect card
          xor  a             ; success
          ret

; disk write
;   input: data in DSKBUF
;  output: A = 0x00 on success
dskwrite: ld   hl, SD_CMD24+1
          call sdlba         ; set LBA address

          ; write command
          call sdon          ; select card
          ld   hl, SD_CMD24  ; write block
          ld   b,  6
          call sdcmd
          jp   c,  dskerr2   ; timeout, error
          and  0xFE          ; any error?
          jp   z,  dskwr1    ; ...no, continue

          ld   hl, SD_MSG9   ; error
          jp   dskerr

dskwr1:   ; write start token
          ld   a, 0xFE
          out  (SPID), a

          ; write sector
          ld   hl, DSKBUF
          ld   bc, 0x0200
dskwr2:   ld   a, (hl)
          out  (SPID), a
          inc  hl
          dec  bc
          ld   a, b
          or   c
          jp   nz, dskwr2

          ; wait for data accepted
dskwr3:   ld   a, 0xFF
          out  (SPID), a     ; send 0xFF
          in   a, (SPID)     ; read data
          cp   0xFF          ; still busy?
          jp   z,  dskwr3    ; ...yes, wait
          and  0x1F
          cp   0x05          ; data accepted?
          jp   z, dskwr4     ; ...yes, continue

          ld   hl, SD_MSG9   ; error
          jp   dskerr

          ; wait for data written
dskwr4:   ld   a, 0xFF
          out  (SPID), a     ; send 0xFF
          in   a, (SPID)     ; read data
          or   a             ; busy?
          jp   z, dskwr4     ; ...yes, wait

          call sdoff         ; deselect card
          xor  a             ; success
          ret

; disk error handlers
dskerr:   push af            ; save error code
          call outstr        ; print message
          pop  af            ; restore error
          call outhex8       ; print it
          ld   hl, SD_CRLF   ; newline
          call outstr
          jp   dskerr3       ; end of handler
dskerr2:  ld   hl, SD_TIME2  ; timeout
          call outstr
          ; fall-through
dskerr3:  call sdoff         ; deselect card
          ld   a, 0xFF       ; no success
          or   a
          ret
